#!/usr/bin/env python3

#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.

#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.

#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

#    evSoc Version 2.3 	is based on EVNnotiPi Release 1.0.0
#    evSoc			Adopted by openWB.de/forum ragnaroek
#
#    tested @ RaspberryPi with Linux Linux 5.10.17-v7l+ 
#           python 3.7.3
#           opwenWB 1.9.220
#    with Kona(2018) using a TONWON OBDII BT3.0 Bluetooth Adapter

from gevent.monkey import patch_all; patch_all()
from subprocess import check_call, check_output
from time import sleep,time
import os
import sys
import signal
import sdnotify
import logging
from argparse import ArgumentParser
import paho.mqtt.client as mqtt

import subprocess

################################# Bluetooth Binding #######################################
def bindBT():
    port = format(config['dongle']['port'])
    addr = format(config['dongle']['addr'])
    result=subprocess.run(['rfcomm','bind','/dev/hci0', addr])
    log.info("bindBT passed")
    result.check_returncode()

def releaseBT():
    port = format(config['dongle']['port'])
    result=subprocess.run(['rfcomm', 'release ','/dev/hci0'])
    log.info("releaseBT passed")
    result.check_returncode()
    
############################## Set up signal handling ######################################
def exit_gracefully(signum, frame):
    sys.exit(0)

#################################### MQTT interface ########################################
def on_connect(client, userdata, flags, rc):
    if (rc != 0):
        log.info("MQTT connect failed with result code " + str(rc))
    else:
        log.info("MQTT subscribed")
    client.subscribe("openWB/lp/1/boolChargeStat")
    client.subscribe("openWB/lp/1/boolPlugStat")
    client.subscribe("evSoc/carport/boolOpenStat")

def on_publish(client, userdata, mid):
   print("published "+userdata)

def on_message(client, userdata, msg):
    global charging
    global pluggedEvent
    if (msg.topic == "openWB/lp/1/boolChargeStat"):
        if (int(msg.payload) == 0):			# Stopped
            log.info("Charging stopped")
            charging = False
        else:
            log.info("Charging started")
            charging = True
    if (msg.topic == "openWB/lp/1/boolPlugStat"):
        if (int(msg.payload) == 0):			# Unplugged
            log.info("Unplugged")
            pluggedEvent = 0
        else:
            log.info("Plugged")
            pluggedEvent = 6
    if (msg.topic == "evSoc/carport/boolOpenStat"):
        if (int(msg.payload) == 0):			# Carport door closed
            log.info("Carport closed")
            pluggedEvent = 0
        else:
            log.info("Carport opened, try to connect with car")
            pluggedEvent = 6

######################################## MAIN ####################################

Systemd = sdnotify.SystemdNotifier()

parser = ArgumentParser(description='evSoc')
parser.add_argument('-d', '--debug', dest='debug', action='store_true', default=False)
parser.add_argument('-c', '--config', dest='config', action='store', default='config.yaml')
args = parser.parse_args()
del parser
    
# load config
if os.path.exists(args.config):
    if args.config[-5:] == '.json':
        import json
        with open(args.config, encoding='utf-8') as config_file:
            config = json.loads(config_file.read())
    elif args.config[-5:] == '.yaml':
        import yaml
        with open(args.config, encoding='utf-8') as config_file:
            config = None
            # use the last document in config.yaml as config
            for c in yaml.load_all(config_file, Loader=yaml.SafeLoader):
                config = c
    else:
        raise Exception('Unknown config type')
else:
    raise Exception('No config found')

if args.debug:
    logging.basicConfig(level=logging.DEBUG)
elif 'loglevel' in config:
    logging.basicConfig(level=config['loglevel'])
else:
    logging.basicConfig(level=logging.INFO)
log = logging.getLogger("evSoc")

del args

# Load OBD2 interface module
if not "{}.py".format(config['dongle']['type']) in os.listdir('dongles'):
    raise Exception('Unsupported dongle {}'.format(config['dongle']['type']))

# Init ODB2 adapter
sys.path.insert(0, 'dongles')
exec("from {0} import {0} as DONGLE".format(config['dongle']['type']))
sys.path.remove('dongles')

if not "{}.py".format(config['car']['type']) in os.listdir('cars'):
    raise Exception('Unsupported car {}'.format(config['car']['type']))

sys.path.insert(0, 'cars')
exec("from {0} import {0} as CAR".format(config['car']['type']))
sys.path.remove('cars')


charging     = False					# status from openWB LP1
pluggedEvent = 3					# status from openWB LP1, 3 periods polling
period       = 10					# seconds for polling data from car

signal.signal(signal.SIGTERM, exit_gracefully)

mqtt_broker_ip = config['mqtt']['server']
client = mqtt.Client()

client.on_connect=on_connect
client.on_message=on_message
#client.on_publish=on_publish

try:
    client.connect(mqtt_broker_ip, config['mqtt']['port'])
except:
    log.error("MQTT-connect failed")

Systemd.notify("READY=1")

client.loop_start()					# now we monitor the MQTT topics
bind   =  False					# if bluetooth rfcomm0 is bind..ed
car    = 'NIL'						# uninitialized but declared for car object
dongle = 'NIL'						# uninitialized but declared for dongle object


# evSoc has been most probably been restarted, because the BT-connection was not available
# we check once only, if car was already plugged (to charging cable), maybe we get an SoC

try:
    while True:
        if charging or pluggedEvent>0:		# charging active OR just plugged 
            try:
                if not bind:
                    bindBT()						# now 12v batt will be discharged
                    bind=True
                    dongle = DONGLE(config['dongle'])			# Init dongle
                    if(dongle.serial.is_open):
                        log.debug("BT link is open")
                    else:
                        log.debug("BT link still closed, i.e. no car!")	# evSoc will die before reaching this point
                    car = CAR(config['car'], dongle, None)		# Init car
                polled=False
                for repeat in range(2):					# repetitions in case of faults
                    SoC=car.pollData()					# -1 if fault happend (ECUs unpowered in car)
                    polled = (SoC >= 0)
                    if not polled:
                        log.warning("Re-Init Dongle - repeat "+str(repeat))
                        dongle.initDongle()
                        sleep(2)
                    else:
                        log.info("SoC=" + str(SoC))
                        client.publish(config['mqtt']['topic'], round(SoC), qos=0, retain=True)
                        pluggedEvent=0
                if not polled:
                    pluggedEvent-=1
            except(SystemExit):
                dummy=0
        if not charging and bind:
            try:
                dongle.serial.close()
            except:
                dummy=0
            releaseBT()						# now 12v batt will NOT be discharged
            bind=False
        log.debug("pluggedEvent "+str(pluggedEvent))
        sleep(period)
except (KeyboardInterrupt, SystemExit): #when you press ctrl+c
    main_running = False
    Systemd.notify("STOPPING=1")
finally:
    if( bind ):
        try:
            dongle.serial.close()
        except:
            dummy=0
        releaseBT()
    Systemd.notify("STOPPING=1")
    client.disconnect()
    log.info("Exiting ...")
    log.info("Bye.")

